//mtexdump
//by thakis
//v1.0

//converts a metroid txtr file to a bti file
//(bti was chosen as output format because
//it was the least work for me ;-) )
#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <cassert>
#include <vector>
#include <sstream>
#include <string.h>

#include "GCCommon.h"

using namespace std;

#pragma pack(push, 1)
struct BtiTextureHeader //for output
{
	u8 format;  //seems to match tpl's format (see yagcd)
	u8 unknown;
	u16 width;
	u16 height;
	u16 unknown2; //probably padding

	u8 unknown3;
	u8 paletteFormat; //matches tpl palette format
	u16 paletteNumEntries;
	u32 paletteOffset;

	u32 unknown5;
	u16 unknown6;
	u16 unknown7;
	u8 mipmapCount;
	u8 unknown8;
	u16 unknown9;

	//this offset is relative to the TextureHeader (?)
	u32 dataOffset;

	//some of the unknown data could be render state?
	//(lod bias, min/mag filter, clamp s/t, ...)
};

struct TxtrHeader
{
	//0 - i4
	//1 - i8
	//2 - i4a4
	//3 - i8a8
	//4 - idx4
	//5 - idx8
	//(there's no file with format 6)
	//7 - r5g6b5
	//8 - rgb5a3
	//9 - r8g8b8a8
	//10 - dxt1
	u32 format;
	u16 w, h;
	u32 numMips; //could be two u16s as well...
};

struct PaletteHeader
{
	//0 - i8a8
	//1 - r5g6b5
	//2 - rgb5a3
	u32 format;

	//unk is 256 for idx8 images and 1 for idx4 images
	//unk2 is 1 for idx8 images and 16 for idx4 images
	u16 unk, unk2;
};
#pragma pack(pop)

int dumpTxtr(FILE* srcFile, string dstFile)
{
    FILE* f = srcFile;
	if(f){
		BtiTextureHeader outH;
		memset(&outH, 0, sizeof(outH));

		TxtrHeader h;
		fread(&h, sizeof(h), 1, f);
		toDWORD(h.format);
		toWORD(h.w);
		toWORD(h.h);
		toDWORD(h.numMips);

		outH.width = h.w;
		outH.height = h.h;
		outH.mipmapCount = h.numMips;
		outH.dataOffset = sizeof(outH);

		//read palette header
		if(h.format == 4 || h.format == 5)
		{
			PaletteHeader ph;
			fread(&ph, sizeof(ph), 1, f);
			toDWORD(ph.format);
			toWORD(ph.unk);
			toWORD(ph.unk2);

			outH.paletteFormat = ph.format;

			//TODO...
			outH.paletteNumEntries = h.format == 4 ? 16 : 256;
			outH.paletteOffset = outH.dataOffset;

			int paletteSize = outH.paletteNumEntries*2;
			outH.dataOffset += paletteSize;
		}

		toWORD(outH.width);
		toWORD(outH.height);
		toDWORD(outH.dataOffset);
		toWORD(outH.paletteNumEntries);
		toDWORD(outH.paletteOffset);

		//set output format
		int size = 0;
		switch(h.format)
		{
		case 0:
			outH.format = 0;
			size = h.w*h.h/2;
			break;

		case 1:
			outH.format = 1;
			size = h.w*h.h;
			break;

		case 2:
			outH.format = 2;
			size = h.w*h.h;
			break;

		case 3:
			outH.format = 3;
			size = h.w*h.h*2;
			break;

		case 4: //TODO, doesn't work
			outH.format = 0;
			size = h.w*h.h/2;
			break;

		case 5:
			outH.format = 9;
			size = h.w*h.h;
			break;

		case 7:
			outH.format = 4;
			size = h.w*h.h*2;
			break;

		case 8:
			outH.format = 5;
			size = h.w*h.h*2;
			break;

		case 9:
			outH.format = 6;
			size = h.w*h.h*4;
			break;

		case 10:
			outH.format = 14;
			size = h.w*h.h/2;
			break;
		}

		//compute image data size
		int imageSize = 0;
		for(int i = 0, s = 1; i < h.numMips; ++i, s *= 2)
			imageSize += size/(s*s);

		FILE* out = fopen(dstFile.c_str(), "wb");
		if(out == NULL)
			return 1;

		fwrite(&outH, sizeof(outH), 1, out);
		while(!feof(f))
			fputc(fgetc(f), out);

		//after the dxt image data, an image footer is stored
		//in the file...let's don't care about it for now

		fclose(out);
	}else{
		return 1;//Errored
	}
	return 0;
}

long RFL(FILE* fp){//Return File Length
	long curroffset=ftell(fp);
	long size=0;
	fseek(fp,0,SEEK_END);
	size=ftell(fp);
	fseek(fp,curroffset,SEEK_SET);
	return size;
}

int main(int argc, char* argv[])
{
  FILE* f;
  if(argc < 2 || (f = fopen(argv[1], "rb")) == NULL)
    return 1;

  string outFilename = argv[1] + string(".bti");
  dumpTxtr(f, outFilename);

  fclose(f);

  return 0;
}
